package com.serval.polsplitter;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import com.serval.polsplitter.structs.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.*;


public class PolicySplitter {


    public static final Subject ANY_SUBJECT = new Subject();
    public static final Action ANY_ACTION = new Action();
    public static final Resource ANY_RESOURCE = new Resource();

    static {

        ANY_SUBJECT.getAttributes().add(new Attribute(" - ", "AnySubject"));
        ANY_ACTION.getAttributes().add(new Attribute(" - ", "AnyAction"));
        ANY_RESOURCE.getAttributes().add(new Attribute(" - ", "AnyResource"));
    }

    // some constants
    public static final int SUBJECT = 1,
            ACTION = 2,
            RESOURCE = 3,
            SUBJECT_ACTION = 4,
            SUBJECT_RESOURCE = 5,
            ACTION_RESOURCE = 6,
            SUBJECT_ACTION_RESOURCE = 7;


    private static final String[] element_names = {"", "Subject", "Action", "Resource"};
    private static final String[] ccNames = {"", "S", "A", "R", "SA", "SR", "AR", "SAR"};

    /**
     * splits the policies into several policies based on many combinations
     *
     * @param filename The XACML policy
     * @param ccList   the cc list to consider when generating split the policy
     */
    public static void splitOnePolicy(String filename, Collection<String> ccList) {

        Map<Integer, Collection<Target>> allCCsMap;
        Collection<Target> targets, tatgetsColTmp;
        String decomposedPolicy;
        File result, policyFile;
        String policyName, cleanedFile;

        // extract all targets from policy rules
        targets = extractAllRulesTargets(filename);

        // compute all Clustering Criteria
        allCCsMap = Utils.computeAllClusteringCriteria(targets);

        // split by by clustering with different params
        for (int ccType = 1; ccType <= 7; ccType++) {

            // skip when it is not the chosen CCs
            if (!ccList.contains(ccNames[ccType]))
                continue;

            tatgetsColTmp = allCCsMap.get(ccType);

            System.out.println("generating file for : " + ccNames[ccType]);

            policyFile = new File(filename);

            policyName = policyFile.getName().substring(policyFile.getName().lastIndexOf(File.separator) + 1, policyFile.getName().lastIndexOf("."));

            // create folder for that CCType
            result = new File(policyFile.getParent() + File.separator + policyName + "_splitted" + File.separator + ccNames[ccType]);

            if (!result.exists()) {
                boolean created = result.mkdir();

                if (!created)
                    System.out.println("error in creating " + policyFile.getParent() + File.separator + policyName + "_splitted" + File.separator + ccNames[ccType]);
            }

            for (Target ccTarget : tatgetsColTmp) {

                // decompose then clean file
                decomposedPolicy = splitPoliciesByTarget(filename, ccTarget, ccType, ccTarget.toShordString());
                cleanedFile = cleanPolicyFile(decomposedPolicy);

                // put the CC target in the policy or policySet target
                Utils.updateTargetInPolicyFile(cleanedFile);
            }
        }
    }


    /**
     * splits policies by a given arguments
     *
     * @param filename the policy file
     * @param ccTarget the target containing the cc param (action, subject etc.)
     * @param ccType   the CC type Subject Action ...
     * @param ccName   the clustering criteria short name
     * @return the name of the created xacml policy
     */
    private static String splitPoliciesByTarget(String filename, Target ccTarget, Integer ccType, String ccName) {

        File file;
        Source source;
        Transformer tFormer;
        StreamResult destination;
        String filePol = filename;
        Collection<Node> nodesToBeDeleted;

        nodesToBeDeleted = new ArrayList<Node>();

        try {

            file = new File(filename);

            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(file);
            doc.getDocumentElement().normalize();


            NodeList rules = doc.getElementsByTagName("Rule");

            if (rules.getLength() > 0) {

                for (int s = 0; s < rules.getLength(); s++) {


                    // extract args
                    Element rule = (Element) rules.item(s);

                    // skip rule if empty
                    if (rule.getChildNodes().getLength() > 0) {

                        // rule may not contain target
                        if (rule.getElementsByTagName("Target").getLength() > 0) {

                            Element targetElement = (Element) rule.getElementsByTagName("Target").item(0);

                            // target may be empty too
                            if (targetElement.getChildNodes().getLength() != 0) {

                                for (TargetElement element : ccTarget.getSubjects()) {
                                    nodesToBeDeleted.addAll(extractNodesToRemove(targetElement, element, SUBJECT));
                                }

                                for (TargetElement element : ccTarget.getActions()) {
                                    nodesToBeDeleted.addAll(extractNodesToRemove(targetElement, element, ACTION));
                                }

                                for (TargetElement element : ccTarget.getResources()) {
                                    nodesToBeDeleted.addAll(extractNodesToRemove(targetElement, element, RESOURCE));
                                }
                            }
                        }
                    }
                }
            }


            // deleting rules nodes & creating xml file
            for (Node node : nodesToBeDeleted) {

                node.getParentNode().removeChild(node);
            }

            String policyName = file.getName().substring(file.getName().lastIndexOf(File.separator) + 1, file.getName().lastIndexOf("."));

            // the folder name is like this mypolicy1_splitted/CC
            String folderName = file.getParent() + File.separator + policyName + "_splitted" + File.separator + ccNames[ccType];

            // filter out reserved characters from policy name
            policyName = removeReservedCharsFromString(policyName);
            ccName = removeReservedCharsFromString(ccName);

            filePol = folderName + File.separator + policyName + "_" + ccNames[ccType] + "_" + ccName + "tmp.xml";

            doc.normalize();
            source = new DOMSource(doc);
            PrintWriter pwOut = new PrintWriter(filePol);
            destination = new StreamResult(pwOut);
            tFormer = TransformerFactory.newInstance().newTransformer();
            tFormer.transform(source, destination);
            pwOut.flush();
            pwOut.close();

        } catch (ParserConfigurationException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (TransformerConfigurationException e) {
            e.printStackTrace();
        } catch (TransformerException e) {
            e.printStackTrace();
        }

        return filePol;
    }


    /**
     * Replaces the reserved chars in filename with _
     *
     * @param fileName the file to use
     * @return filename with reserved chars replaced by _
     */
    public static String removeReservedCharsFromString(String fileName) {

        String filter = "<>:\"/\\|?*";

        for (int i = 0; i < filter.length(); i++) {

            fileName = fileName.replace(filter.charAt(i) + "", "-");

        }

        return fileName;
    }

    /**
     * Cleans the policy file by removing rule where target elements are empty
     *
     * @param filename the policy to be cleaned filename
     * @return the filename
     */
    private static String cleanPolicyFile(String filename) {

        File file;
        Source source;
        Transformer tFormer;
        StreamResult dest;
        String filePolicyName = null;
        Collection<Node> nodesToBeDeleted;
        Collection<TargetElement> subjects, actions, resources;

        nodesToBeDeleted = new ArrayList<Node>();

        try {

            file = new File(filename);

            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();

            // System.out.println(filename);
            Document doc = db.parse(file);
            doc.getDocumentElement().normalize();

            NodeList rules = doc.getElementsByTagName("Rule");

            if (rules.getLength() > 0) {

                for (int s = 0; s < rules.getLength(); s++) {

                    // extract args
                    Element rule = (Element) rules.item(s);

                    // skip rule if empty
                    if (rule.getChildNodes().getLength() > 0) {

                        // rule may not contain target
                        if (rule.getElementsByTagName("Target").getLength() > 0) {

                            Element targetElement = (Element) rule.getElementsByTagName("Target").item(0);

                            subjects = extractOneTargetElement(targetElement, SUBJECT);
                            actions = extractOneTargetElement(targetElement, ACTION);
                            resources = extractOneTargetElement(targetElement, RESOURCE);

                            // if one of the lists is emplty rule should be removed
                            if (subjects.isEmpty() || actions.isEmpty() || resources.isEmpty()) {
                                nodesToBeDeleted.add(rule);
                            }
                        }
                    }
                }
            }

            // deleting rules nodes & creating xml file
            for (Node node : nodesToBeDeleted) {
                node.getParentNode().removeChild(node);
            }

            filePolicyName = filename.substring(0, filename.lastIndexOf("tmp") - 1) + ".xml";

            doc.normalize();
            source = new DOMSource(doc);
            dest = new StreamResult(new PrintWriter(filePolicyName));
            tFormer = TransformerFactory.newInstance().newTransformer();
            tFormer.transform(source, dest);

            // remove tmp file
            new File(filename).delete();

        } catch (ParserConfigurationException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (TransformerConfigurationException e) {
            e.printStackTrace();
        } catch (TransformerException e) {
            e.printStackTrace();
        }

        return filePolicyName;
    }


    /**
     * Extracts from the policy rule Targets lists of Subject, Action, Resource and all 2 combinations (Subject,Action) etc. and all 3 combination
     * used in targets
     *
     * @param filename the XACML policy
     * @return the map containing all collections
     */
    public static Collection<Target> extractAllRulesTargets(String filename) {

        // elements
        Collection<TargetElement> subjects, actions, resources;
        Target target;

        Collection<Target> result;

        result = new HashSet<Target>();

        try {
            File file = new File(filename);

            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(file);
            doc.getDocumentElement().normalize();

            NodeList rules = doc.getElementsByTagName("Rule");

            if (rules.getLength() > 0) {

                for (int s = 0; s < rules.getLength(); s++) {


                    // extract args
                    Element rule = (Element) rules.item(s);

                    // skip rule if empty
                    if (rule.getChildNodes().getLength() > 0) {

                        // rule may not contain target
                        if (rule.getElementsByTagName("Target").getLength() > 0) {


                            Element targetElement = (Element) rule.getElementsByTagName("Target").item(0);

                            // target may be empty too
                            if (targetElement.getChildNodes().getLength() != 0) {

                                // System.out.println("Rule " + s + " Target *******************************************************");

                                subjects = extractOneTargetElement(targetElement, SUBJECT);
                                actions = extractOneTargetElement(targetElement, ACTION);
                                resources = extractOneTargetElement(targetElement, RESOURCE);

                                target = new Target(subjects, actions, resources);

                                result.add(target);

                                // System.out.println("**************************************************************************\n\n");

                            }
                        }
                    }
                }
            }


        } catch (ParserConfigurationException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return result;
    }


    /**
     * Extracts all nodes that are different from the element (subject,action;resource)
     *
     * @param target      the rule or policy target
     * @param element     the targetElement (the subject/action/resource)
     * @param elementType is one of the constants SUBJECT , ACTION OR RESOURCE
     * @return the collection targetElements (subjects,actions or resources
     */
    private static Collection<Node> extractNodesToRemove(Element target, TargetElement element, int elementType) {

        Collection<Node> result;
        Element elementsNode, elementMatch, elementAttributeDesignator;
        NodeList elements, elementMatchList;
        int nbElement;
        Node elementNode, attributeValueNode;
        String id, value;

        TargetElement elementObject = null;

        result = new ArrayList<Node>();

        // get the type of the element from the constant array subject, action or resource
        String elemName = element_names[elementType];

        // just ignore the AnySubject/AnyAction/AnyResource
        elementsNode = (Element) target.getElementsByTagName(elemName + "s").item(0);
        if (elementsNode.getElementsByTagName("Any" + elemName).getLength() == 0) {

            elements = elementsNode.getElementsByTagName(elemName);

            nbElement = elements.getLength();

            // read subject by subject
            for (int i = 0; i < nbElement; i++) {

                elementNode = elements.item(i);

                // get elementMatch(s) then for each of them go through to pull Attribute & SubjectAttributeDesignator

                switch (elementType) {
                    case SUBJECT:
                        elementObject = new Subject();
                        break;
                    case ACTION:
                        elementObject = new Action();
                        break;
                    case RESOURCE:
                        elementObject = new Resource();
                        break;
                    default:
                        break;
                }

                elementMatchList = ((Element) elementNode).getElementsByTagName(elemName + "Match");
                for (int j = 0; j < elementMatchList.getLength(); j++) {

                    // pull attribute value &  attributeID SubjectAttributeDesignator from elementMatch item

                    elementMatch = (Element) elementMatchList.item(j);

                    attributeValueNode = elementMatch.getElementsByTagName("AttributeValue").item(0);
                    elementAttributeDesignator = (Element) elementMatch.getElementsByTagName(elemName + "AttributeDesignator").item(0);

                    id = elementAttributeDesignator.getAttribute("AttributeId");
                    value = attributeValueNode.getFirstChild().getNodeValue();

                    // System.out.println("For " + elemName + " number " + (i + 1) + "   ->  attribute ID : " + id + " , value = " + value);

                    if (elementObject != null) {
                        elementObject.getAttributes().add(new Attribute(id, value));
                    }
                }

                // add to result if different from the target element
                if (!element.equals(elementObject))
                    result.add(elementNode);


            }


        }
        return result;
    }


    /**
     * Extracts all target element of a given type (subject, action or resource from a given target)
     *
     * @param target      the rule or policy target
     * @param elementType is one of the constants SUBJECT , ACTION OR RESOURCE
     * @return the collection targetElements (subjects,actions or resources
     */
    private static Collection<TargetElement> extractOneTargetElement(Element target, int elementType) {

        Collection<TargetElement> result;
        Element elementsNode, elementMatch, elementAttributeDesignator;
        NodeList elements, elementMatchList;
        int nbElement;
        Node elementNode, attributeValueNode;
        String id, value;

        TargetElement elementObject = null;

        result = new ArrayList<TargetElement>();

        // get the type of the element from the constant array subject, action or resource
        String elemName = element_names[elementType];

        elementsNode = (Element) target.getElementsByTagName(elemName + "s").item(0);

        // if it is AnySubject or AnyResource or AnyAction just add it
        if (elementsNode.getElementsByTagName("Any" + elemName).getLength() > 0) {

            switch (elementType) {
                case SUBJECT:
                    result.add(ANY_SUBJECT);
                    break;
                case ACTION:
                    result.add(ANY_ACTION);
                    break;
                case RESOURCE:
                    result.add(ANY_RESOURCE);
                    break;
                default:
                    break;
            }

            // System.out.println("For " + elemName + " attribute ID : - , value = Any" + elemName);

        } else {

            elements = elementsNode.getElementsByTagName(elemName);

            nbElement = elements.getLength();

            // read subject by subject
            for (int i = 0; i < nbElement; i++) {

                elementNode = elements.item(i);

                // get elementMatch(s) then for each of them go through to pull Attribute & SubjectAttributeDesignator

                switch (elementType) {
                    case SUBJECT:
                        elementObject = new Subject();
                        break;
                    case ACTION:
                        elementObject = new Action();
                        break;
                    case RESOURCE:
                        elementObject = new Resource();
                        break;

                    default:
                        break;
                }

                elementMatchList = ((Element) elementNode).getElementsByTagName(elemName + "Match");

                for (int j = 0; j < elementMatchList.getLength(); j++) {

                    // pull attribute value &  attributeID SubjectAttributeDesignator from elementMatch item

                    elementMatch = (Element) elementMatchList.item(j);

                    attributeValueNode = elementMatch.getElementsByTagName("AttributeValue").item(0);
                    elementAttributeDesignator = (Element) elementMatch.getElementsByTagName(elemName + "AttributeDesignator").item(0);

                    id = elementAttributeDesignator.getAttribute("AttributeId");
                    value = attributeValueNode.getFirstChild().getNodeValue();

                    // System.out.println("For " + elemName + " number " + (i + 1) + "   ->  attribute ID : " + id + " , value = " + value);

                    if (elementObject != null) {
                        elementObject.getAttributes().add(new Attribute(id, value));
                    }
                }
                // add to result
                result.add(elementObject);
            }
        }

        return result;
    }

    /**
     * splits policies and puts them in results/ folder
     * l  *
     *
     * @param args   the files
     * @param ccList the cc list to consider when generating split policies
     */
    public static void splitPoliciesInCollection(Collection<String> args, Collection<String> ccList) {

        File policyFile;
        boolean created;
        String policyName;

        for (String filename : args) {

            System.out.println("Handling file: " + filename + ";");
            System.out.println("SC type; # PDPs;");


            // create the result folder
            policyFile = new File(filename);
            policyName = policyFile.getName().substring(policyFile.getName().lastIndexOf(File.separator) + 1, policyFile.getName().lastIndexOf("."));

            File result = new File(policyFile.getParent() + File.separator + policyName + "_splitted");

            if (!result.exists()) {
                created = result.mkdir();

                if (!created) System.out.println("error in creating result folder");
            }

            System.out.println("Processing file : " + filename);
            splitOnePolicy(filename, ccList);
        }
    }


}